import inspect
from typing import Any, Callable, List, Optional, get_type_hints

from unfazed.utils import Storage

from unfazed_taskiq.schema.registry.task import RegistryTaskParam, RegistryTaskSchema


class RegistryTask(Storage[RegistryTaskSchema]):
    def _register(self, path: str, task: RegistryTaskSchema) -> None:
        if path in self.storage:
            raise ValueError(f"Task {path} already registered")
        self.storage[path] = task

    def get(self, path: str) -> Optional[RegistryTaskSchema]:
        return self.storage.get(path, None)

    def filter_path(self, keyword: Optional[str] = None) -> List[RegistryTaskSchema]:
        res: List[RegistryTaskSchema] = []
        for k, v in self.storage.items():
            if keyword is None or keyword in k:
                res.append(v)
        return res

    def register_broker(
        self,
        func: Callable,
        alias_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Any:
        params_info: list[RegistryTaskParam] = []
        sig = inspect.signature(func)
        type_hints = get_type_hints(func)
        for name, param in sig.parameters.items():
            param_type = type_hints.get(name, Any)
            required = param.default is inspect.Parameter.empty
            default = None if required else param.default

            params_info.append(
                RegistryTaskParam(
                    **{
                        "name": name,
                        "hint_type": param_type,
                        "required": required,
                        "default": default,
                    }
                )
            )
        registry_task: RegistryTaskSchema = RegistryTaskSchema(
            name=func.__name__,
            alias_name=alias_name,
            params=params_info,
            docs=func.__doc__ or "",
            schedule=kwargs.get("schedule", None),
        )

        task_path = f"{func.__module__}.{func.__name__}"

        self._register(task_path, registry_task)


rs = RegistryTask()



import importlib
import sys
from types import ModuleType, SimpleNamespace
from typing import Any, Dict
from unittest.mock import AsyncMock, MagicMock, patch

import pytest


@pytest.fixture
def handler_module(monkeypatch: pytest.MonkeyPatch) -> ModuleType:
    alias = "test_alias"

    def fake_import_setting(key: str) -> Dict[str, Dict[str, object]]:
        assert key == "UNFAZED_SETTINGS_MODULE"
        return {
            "UNFAZED_TASKIQ_SETTINGS": {
                "DEFAULT_TASKIQ_NAME": alias,
                "TASKIQ_CONFIG": {
                    alias: {
                        "BROKER": {
                            "BACKEND": "taskiq.InMemoryBroker",
                            "OPTIONS": {},
                        },
                    }
                },
            }
        }

    monkeypatch.setenv("UNFAZED_SETTINGS_MODULE", "tests.proj.entry.settings")
    monkeypatch.setattr("unfazed.utils.import_setting", fake_import_setting)

    def fake_setup(setup_alias: str, config: object) -> SimpleNamespace:
        return SimpleNamespace(
            alias_name=setup_alias,
            broker=f"broker-{setup_alias}",
            scheduler=f"scheduler-{setup_alias}",
            startup=AsyncMock(name=f"startup-{setup_alias}"),
            shutdown=AsyncMock(name=f"shutdown-{setup_alias}"),
        )

    monkeypatch.setattr("unfazed_taskiq.agent.model.TaskiqAgent.setup", fake_setup)

    sys.modules.pop("unfazed_taskiq.agent.handler", None)
    module = importlib.import_module("unfazed_taskiq.agent.handler")
    return module


class TestAgentHandler:
    def _make_handler(self, module: Any, monkeypatch: pytest.MonkeyPatch) -> Any:
        handler_cls = module.AgentHandler
        with patch.object(handler_cls, "check_ready", autospec=True) as mock_ready:
            mock_ready.side_effect = lambda self: None
            handler = handler_cls()
        return handler

    def test_module_singletons(self, handler_module: Any) -> None:
        alias = handler_module.agents.default_alias_name
        assert (
            handler_module.scheduler == handler_module.agents.storage[alias].scheduler
        )
        assert handler_module.broker == handler_module.agents.storage[alias].broker

    def test_init_triggers_check_ready(self, handler_module: Any) -> None:
        handler_cls = handler_module.AgentHandler
        with patch.object(handler_cls, "check_ready", autospec=True) as mock_ready:
            handler = handler_cls()
        mock_ready.assert_called_once()
        assert isinstance(handler, handler_cls)

    def test_register_and_duplicate(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        agents = MagicMock()
        handler.register("alpha", agents)
        assert handler.storage["alpha"] is agents
        with pytest.raises(ValueError, match="already registered"):
            handler.register("alpha", agents)

    def test_reset(self, handler_module: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler.register("alpha", MagicMock())
        handler.reset()
        assert handler.storage == {}
        assert handler._ready is False

    def test_setup_requires_env(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        monkeypatch.delenv("UNFAZED_SETTINGS_MODULE", raising=False)
        with pytest.raises(ValueError, match="is not set"):
            handler.setup()

    def test_setup_import_error(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        monkeypatch.setenv("UNFAZED_SETTINGS_MODULE", "module.path")
        monkeypatch.setattr(
            handler_module,
            "import_setting",
            MagicMock(side_effect=ImportError("boom")),
        )
        with pytest.raises(ImportError, match="Failed to import settings module: boom"):
            handler.setup()

    def test_setup_invalid_config(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        monkeypatch.setenv("UNFAZED_SETTINGS_MODULE", "module.path")
        monkeypatch.setattr(
            handler_module,
            "import_setting",
            lambda key: {"UNFAZED_TASKIQ_SETTINGS": {"DEFAULT_TASKIQ_NAME": "alpha"}},
        )

        def bad_validate(cls: Any, values: Any) -> Any:
            raise ValueError("bad config")

        monkeypatch.setattr(
            handler_module.UnfazedTaskiqSettings,
            "model_validate",
            classmethod(bad_validate),
        )

        with pytest.raises(
            ValueError, match="Invalid settings configuration: bad config"
        ):
            handler.setup()

    def test_setup_success(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        monkeypatch.setenv("UNFAZED_SETTINGS_MODULE", "module.path")
        config = {
            "DEFAULT_TASKIQ_NAME": "alpha",
            "TASKIQ_CONFIG": {
                "alpha": {
                    "BROKER": {
                        "BACKEND": "taskiq.InMemoryBroker",
                        "OPTIONS": {},
                    }
                }
            },
        }
        monkeypatch.setattr(
            handler_module,
            "import_setting",
            lambda _: {"UNFAZED_TASKIQ_SETTINGS": config},
        )

        fake_agent = SimpleNamespace(
            alias_name="alpha",
            broker="broker-alpha",
            scheduler="scheduler-alpha",
            startup=AsyncMock(),
            shutdown=AsyncMock(),
        )

        monkeypatch.setattr(
            handler_module.TaskiqAgent,
            "setup",
            MagicMock(return_value=fake_agent),
        )

        handler.setup()

        assert handler.default_alias_name == "alpha"
        assert handler.storage["alpha"] is fake_agent
        assert handler._ready is True

    def test_check_ready_paths(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        with (
            patch.object(handler, "reset") as mock_reset,
            patch.object(handler, "setup") as mock_setup,
        ):
            handler._ready = False
            handler.check_ready()
        mock_reset.assert_called_once()
        mock_setup.assert_called_once()

        with (
            patch.object(handler, "reset") as mock_reset,
            patch.object(handler, "setup") as mock_setup,
        ):
            handler._ready = True
            handler.check_ready()
        mock_reset.assert_not_called()
        mock_setup.assert_not_called()

    def test_get_agent_behaviour(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler.default_alias_name = "alpha"
        sentinel_agents = MagicMock()
        handler.storage["alpha"] = sentinel_agents
        handler._ready = True
        assert handler.get_agent(None) is sentinel_agents
        assert handler.get_agent("missing") is None

    def test_get_agent_triggers_ready(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler._ready = False
        with patch.object(handler, "check_ready") as mock_ready:
            handler.get_agent("alpha")
        mock_ready.assert_called_once()

    def test_scheduler_property(
        self, handler_module: Any, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        node = SimpleNamespace(scheduler="scheduler-alpha")
        handler.storage["alpha"] = node
        handler.default_alias_name = "alpha"
        handler._ready = True
        assert handler.scheduler == "scheduler-alpha"

    def test_scheduler_triggers_ready(
        self, handler_module: ModuleType, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler.storage["alpha"] = SimpleNamespace(scheduler="S")
        handler.default_alias_name = "alpha"
        handler._ready = False

        def mark_ready() -> None:
            handler._ready = True

        with patch.object(handler, "check_ready", side_effect=mark_ready) as mock_ready:
            assert handler.scheduler == "S"
        mock_ready.assert_called_once()

    def test_broker_property(
        self, handler_module: ModuleType, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler.storage["alpha"] = SimpleNamespace(broker="broker-alpha")
        handler.default_alias_name = "alpha"
        handler._ready = True
        assert handler.broker == "broker-alpha"

    def test_broker_triggers_ready(
        self, handler_module: ModuleType, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        handler.storage["alpha"] = SimpleNamespace(broker="B")
        handler.default_alias_name = "alpha"
        handler._ready = False

        def mark_ready() -> None:
            handler._ready = True

        with patch.object(handler, "check_ready", side_effect=mark_ready) as mock_ready:
            assert handler.broker == "B"
        mock_ready.assert_called_once()

    @pytest.mark.asyncio
    async def test_startup_iterates_agents(
        self, handler_module: ModuleType, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        first = SimpleNamespace(startup=AsyncMock(), shutdown=AsyncMock())
        second = SimpleNamespace(startup=AsyncMock(), shutdown=AsyncMock())
        handler.storage = {"a": first, "b": second}

        await handler.startup()

        first.startup.assert_awaited_once()
        second.startup.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_shutdown_iterates_agents(
        self, handler_module: ModuleType, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        handler = self._make_handler(handler_module, monkeypatch)
        first = SimpleNamespace(startup=AsyncMock(), shutdown=AsyncMock())
        second = SimpleNamespace(startup=AsyncMock(), shutdown=AsyncMock())
        handler.storage = {"a": first, "b": second}

        await handler.shutdown()

        first.shutdown.assert_awaited_once()
        second.shutdown.assert_awaited_once()



import logging
import os
import sys
import typing as t
import uuid
from datetime import datetime

import pytest
from pytest import Item
from pytest_asyncio import is_async_test
from tortoise import Tortoise
from tortoise.connection import connections
from unfazed.core import Unfazed

from unfazed_taskiq.contrib.scheduler.models import PeriodicTask

logger = logging.getLogger("unfazed.taskiq")


# dont need decorate test functions with pytest.mark.asyncio
def pytest_collection_modifyitems(items: t.List[Item]) -> None:
    pytest_asyncio_tests = (item for item in items if is_async_test(item))
    function_scope_marker = pytest.mark.asyncio(loop_scope="function")
    for async_test in pytest_asyncio_tests:
        async_test.add_marker(function_scope_marker, append=False)


# create a global unfazed
# use this fixture in your test functions
@pytest.fixture(autouse=True, scope="function")
async def unfazed() -> t.AsyncGenerator[Unfazed, None]:
    # Clear task registry before each test
    from unfazed_taskiq.agent import agents
    from unfazed_taskiq.registry.task import rs

    rs.clear()
    agents.reset()
    agents.check_ready()

    root_path = os.path.dirname(os.path.abspath(__file__))
    sys.path.append(root_path)
    os.environ.setdefault("UNFAZED_SETTINGS_MODULE", "tests.proj.entry.settings")
    unfazed = Unfazed()
    await unfazed.setup()

    conn = connections.get("default")
    if os.getenv("DISABLE_DB_CREATE", "false") == "false":
        await conn.db_create()
    logger.info("Database created")
    await Tortoise.generate_schemas()
    logger.info("Schemas generated")

    yield unfazed

    await Tortoise._drop_databases()


@pytest.fixture()
async def test_scheduler_sample_data() -> t.Any:
    data: list[dict] = [
        {
            "description": "test_task1",
            "task_args": "[]",
            "task_name": "test.tasks:test_task1",
            "schedule_alias": "test_schedule1",
            "task_kwargs": '{"category": "00001"}',
            "labels": "{}",
            "cron": "* * * * *",
            "total_run_count": 0,
            "enabled": 0,
            "last_run_at": datetime.now(),
            "time": datetime.now(),
            "updated_at": datetime.now(),
            "created_at": datetime.now(),
            "schedule_id": str(uuid.uuid4().hex),
        },
        {
            "description": "test_task2",
            "task_args": "[]",
            "task_name": "test.tasks:test_task2",
            "schedule_alias": "test_schedule2",
            "task_kwargs": '{"keyword": "三角洲行动"}',
            "labels": "{}",
            "cron": "* * * * *",
            "total_run_count": 0,
            "enabled": 0,
            "last_run_at": datetime.now(),
            "time": datetime.now(),
            "updated_at": datetime.now(),
            "created_at": datetime.now(),
            "schedule_id": str(uuid.uuid4().hex),
        },
        {
            "description": "test_task3",
            "task_args": "[]",
            "task_name": "test.tasks:test_task3",
            "schedule_alias": "test_schedule3",
            "task_kwargs": '{"keyword": "三角洲行动"}',
            "labels": "{}",
            "cron": "* * * * *",
            "total_run_count": 0,
            "enabled": 1,
            "last_run_at": datetime.now(),
            "time": datetime.now(),
            "updated_at": datetime.now(),
            "created_at": datetime.now(),
            "schedule_id": str(uuid.uuid4().hex),
        },
        {
            "description": "test_task4",
            "task_args": "[]",
            "task_name": "test.tasks:test_task4",
            "schedule_alias": "test_schedule4",
            "task_kwargs": '{"keyword": "三角洲行动"}',
            "labels": "{}",
            "cron": "* * * * *",
            "total_run_count": 0,
            "enabled": 0,
            "last_run_at": datetime.now(),
            "time": datetime.now(),
            "updated_at": datetime.now(),
            "created_at": datetime.now(),
            "schedule_id": str(uuid.uuid4().hex),
        },
    ]

    for item in data:
        await PeriodicTask.create(**item)

    yield data

    await PeriodicTask.filter().delete()



from typing import Any, Optional

from pydantic import BaseModel


class RegistryTaskParam(BaseModel):
    name: str
    hint_type: Any
    required: bool
    default: Any


class RegistryTaskSchema(BaseModel):
    name: str
    alias_name: Optional[str]
    params: list[RegistryTaskParam]
    docs: Optional[str]
    schedule: Optional[list[dict[str, Any]]]



from unfazed.core import Unfazed
from unfazed.lifespan import BaseLifeSpan

from unfazed_taskiq.agent.handler import agents


class TaskiqLifeSpan(BaseLifeSpan):
    def __init__(self, unfazed: Unfazed) -> None:
        self.unfazed = unfazed
        agents.setup()
        self.agents = agents

    async def on_startup(self) -> None:
        await self.agents.startup()

    async def on_shutdown(self) -> None:
        await self.agents.shutdown()



import json
from typing import Optional

from taskiq import ScheduledTask

from unfazed_taskiq.contrib.scheduler.models import PeriodicTask
from unfazed_taskiq.contrib.scheduler.serializer import PeriodicTaskSerializer


class TestPeriodicTaskModel(object):
    async def test_periodic_task_model(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        assert test_scheduler_sample_data is not None
        assert len(test_scheduler_sample_data) == 4

        db_data = await PeriodicTask.filter().all()

        assert len(db_data) == 4

        db_task_names = {item.task_name for item in db_data}
        sample_task_names = {item["task_name"] for item in test_scheduler_sample_data}
        assert db_task_names == sample_task_names

        db_task_corns = {item.cron for item in db_data}
        sample_task_corns = {item["cron"] for item in test_scheduler_sample_data}
        assert db_task_corns == sample_task_corns

        db_task_kwargs = {item.task_kwargs for item in db_data}
        sample_task_kwargs = {
            item["task_kwargs"] for item in test_scheduler_sample_data
        }
        assert db_task_kwargs == sample_task_kwargs

        db_task_schedule_aliases = {item.schedule_alias for item in db_data}
        sample_task_schedule_aliases = {
            item["schedule_alias"] for item in test_scheduler_sample_data
        }
        assert db_task_schedule_aliases == sample_task_schedule_aliases

    async def test_periodic_task_model_to_taskiq_schedule_task(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        assert test_scheduler_sample_data is not None
        assert len(test_scheduler_sample_data) == 4

        for item in test_scheduler_sample_data:
            pt_objs: Optional[PeriodicTask] = await PeriodicTask.filter(
                task_name=item["task_name"]
            ).first()
            assert pt_objs is not None
            schedule_data: ScheduledTask = pt_objs.to_taskiq_schedule_task()
            assert schedule_data.cron == item["cron"]
            assert schedule_data.task_name == item["task_name"]
            assert schedule_data.args == json.loads(item["task_args"])
            assert schedule_data.kwargs == json.loads(item["task_kwargs"])
            assert schedule_data.labels == json.loads(item["labels"])
            assert schedule_data.schedule_id == item["schedule_id"]

    async def test_periodic_task_model_to_taskiq_schedule_task_with_time(self) -> None:
        """Test to_taskiq_schedule_task method with time-based scheduling."""
        from datetime import datetime, timezone

        # Create a PeriodicTask with time-based scheduling (no cron)
        task_time = datetime(2024, 1, 1, 12, 0, 0, tzinfo=timezone.utc)

        periodic_task = await PeriodicTask.create(
            task_name="test.time_task",
            task_args="[]",
            task_kwargs="{}",
            labels="{}",
            time=task_time,  # Set time instead of cron
            cron=None,  # Explicitly set cron to None
        )

        # Test the conversion
        scheduled_task = periodic_task.to_taskiq_schedule_task()

        # Verify the scheduled task properties
        assert scheduled_task.task_name == "test.time_task"
        assert scheduled_task.args == []
        assert scheduled_task.kwargs == {}
        assert scheduled_task.labels == {}
        assert scheduled_task.time == task_time
        assert scheduled_task.cron is None
        assert scheduled_task.schedule_id == periodic_task.schedule_id

    async def test_periodic_task_model_to_taskiq_schedule_task_no_schedule_error(
        self,
    ) -> None:
        """Test to_taskiq_schedule_task method raises RuntimeError when no schedule is found."""
        import pytest

        # Create a PeriodicTask with neither cron nor time
        periodic_task = await PeriodicTask.create(
            task_name="test.no_schedule_task",
            task_args="[]",
            task_kwargs="{}",
            labels="{}",
            cron=None,
            time=None,
        )

        # Test that RuntimeError is raised
        with pytest.raises(RuntimeError, match="No schedule found"):
            periodic_task.to_taskiq_schedule_task()


class TestPeriodicTaskSerializer(object):
    async def test_periodic_task_serializer(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        assert test_scheduler_sample_data is not None
        assert len(test_scheduler_sample_data) == 4

        result = await PeriodicTaskSerializer.list(
            PeriodicTask.filter(), page=1, size=2
        )
        assert result is not None
        assert len(result.data) == 2



import asyncio
from dataclasses import replace
from typing import Sequence

from taskiq.abc.cmd import TaskiqCMD
from taskiq.cli.scheduler.run import run_scheduler
from unfazed.core import Unfazed

from unfazed_taskiq.agent.handler import agents
from unfazed_taskiq.cli.scheduler.args import SchedulerEventArgs


class SchedulerCMD(TaskiqCMD):
    """Command for taskiq scheduler."""

    short_help = "Run task scheduler"

    async def init_unfazed(self) -> None:
        self.unfazed = Unfazed(silent=True)
        await self.unfazed.setup()

    def exec(self, args: Sequence[str]) -> None:
        """
        Run task scheduler.

        This function starts scheduler function.

        It periodically loads schedule for tasks
        and executes them.

        :param args: CLI arguments.
        """

        # setup unfazed
        asyncio.run(self.init_unfazed())
        # setup scheduler
        parsed: SchedulerEventArgs = SchedulerEventArgs.from_cli(args)

        async def _run_all_scheduler() -> None:
            tasks = []

            # Get all agent models with schedulers
            schedulers = {}
            for alias, agent_model in agents.storage.items():
                if agent_model.scheduler is not None:
                    schedulers[alias] = agent_model.scheduler

            # if alias_name is not provided, run all schedulers
            if len(parsed.alias_name) == 0:
                parsed.alias_name = list(schedulers.keys())

            # init all schedulers
            for alias, scheduler_obj in schedulers.items():
                if alias in parsed.alias_name:
                    event_parsed = replace(parsed, scheduler=scheduler_obj)
                    tasks.append(asyncio.create_task(run_scheduler(event_parsed)))

            # run all schedulers
            if tasks:
                await asyncio.gather(*tasks)

        asyncio.run(_run_all_scheduler())



import json
import uuid
from typing import Optional

from taskiq import ScheduledTask
from tortoise import Tortoise

from unfazed_taskiq.contrib.scheduler.models import PeriodicTask
from unfazed_taskiq.contrib.scheduler.sources import TortoiseScheduleSource


def test_startup_handler() -> None:
    pass


async def test_async_startup_handler() -> None:
    pass


def test_shutdown_handler() -> None:
    pass


async def test_async_shutdown_handler() -> None:
    pass


class TestTortoiseScheduleSource(object):
    async def test_tortoise_schedule_source_init(self) -> None:
        """Test TortoiseScheduleSource initialization."""
        source = TortoiseScheduleSource(schedule_alias="test_task3")
        assert source is not None
        assert source.schedule_alias == "test_task3"
        assert source._alias == "default"
        assert source.startup_handlers == []
        assert source.shutdown_handlers == []

    async def test_tortoise_schedule_source_startup(self) -> None:
        """Test TortoiseScheduleSource startup."""
        source = TortoiseScheduleSource(
            schedule_alias="test_task3",
            startup_handlers=[
                "tests.test_scheduler.test_source.test_startup_handler",
                "tests.test_scheduler.test_source.test_async_startup_handler",
            ],
        )
        await source.startup()
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

    async def test_tortoise_schedule_source_shutdown(self) -> None:
        """Test TortoiseScheduleSource shutdown."""
        source = TortoiseScheduleSource(
            schedule_alias="test_task3",
            shutdown_handlers=[
                "tests.test_scheduler.test_source.test_shutdown_handler",
                "tests.test_scheduler.test_source.test_async_shutdown_handler",
            ],
        )
        await source.startup()
        await source.shutdown()
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

    async def test_tortoise_schedule_source_get_schedules(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        """Test TortoiseScheduleSource get schedules."""
        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()

        assert len(test_scheduler_sample_data) == 4

        simple_data = list(
            filter(
                lambda x: x["schedule_alias"] == "test_schedule3",
                test_scheduler_sample_data,
            )
        )
        schedules = await source.get_schedules()

        assert len(schedules) == 1
        assert schedules[0].task_name == simple_data[0]["task_name"]
        assert schedules[0].args == json.loads(simple_data[0]["task_args"])
        assert schedules[0].kwargs == json.loads(simple_data[0]["task_kwargs"])
        assert schedules[0].labels == json.loads(simple_data[0]["labels"])
        assert schedules[0].schedule_id == simple_data[0]["schedule_id"]

    async def test_tortoise_schedule_source_add_schedule(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        """Test TortoiseScheduleSource add schedule."""
        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()
        s_task = ScheduledTask(
            task_name=test_scheduler_sample_data[0]["task_name"],
            args=json.loads(test_scheduler_sample_data[0]["task_args"]),
            kwargs=json.loads(test_scheduler_sample_data[0]["task_kwargs"]),
            labels=json.loads(test_scheduler_sample_data[0]["labels"]),
            schedule_id=str(uuid.uuid4().hex),
            cron=test_scheduler_sample_data[0]["cron"],
        )
        await source.add_schedule(s_task)
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

        schedules = await source.get_schedules()
        assert len(schedules) == 2

    async def test_tortoise_schedule_source_delete_schedule(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        """Test TortoiseScheduleSource delete schedule."""
        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()

        result_scheduler = list(
            filter(
                lambda x: x["schedule_alias"] == "test_schedule3",
                test_scheduler_sample_data,
            )
        )
        await source.delete_schedule(result_scheduler[0]["schedule_id"])
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

        schedules = await source.get_schedules()
        assert len(schedules) == 0

    async def test_tortoise_schedule_source_pre_send(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        """Test TortoiseScheduleSource pre send."""
        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()
        result_scheduler = list(
            filter(
                lambda x: x["schedule_alias"] == "test_schedule3",
                test_scheduler_sample_data,
            )
        )
        s_task = ScheduledTask(
            task_name=result_scheduler[0]["task_name"],
            args=json.loads(result_scheduler[0]["task_args"]),
            kwargs=json.loads(result_scheduler[0]["task_kwargs"]),
            labels=json.loads(result_scheduler[0]["labels"]),
            schedule_id=result_scheduler[0]["schedule_id"],
            cron=result_scheduler[0]["cron"],
        )
        origin_db_data = await PeriodicTask.filter(
            schedule_id=s_task.schedule_id
        ).first()

        await source.pre_send(s_task)
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

        db_data: Optional[PeriodicTask] = await PeriodicTask.filter(
            schedule_id=s_task.schedule_id
        ).first()
        assert db_data is not None
        assert origin_db_data.last_run_at != db_data.last_run_at  # type: ignore

    async def test_tortoise_schedule_source_post_send(
        self, test_scheduler_sample_data: list[dict]
    ) -> None:
        """Test TortoiseScheduleSource post send."""
        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()
        result_scheduler = list(
            filter(
                lambda x: x["schedule_alias"] == "test_schedule3",
                test_scheduler_sample_data,
            )
        )
        s_task = ScheduledTask(
            task_name=result_scheduler[0]["task_name"],
            args=json.loads(result_scheduler[0]["task_args"]),
            kwargs=json.loads(result_scheduler[0]["task_kwargs"]),
            labels=json.loads(result_scheduler[0]["labels"]),
            schedule_id=result_scheduler[0]["schedule_id"],
            cron=result_scheduler[0]["cron"],
        )

        origin_db_data: Optional[PeriodicTask] = await PeriodicTask.filter(
            schedule_id=s_task.schedule_id
        ).first()
        assert origin_db_data is not None

        await source.post_send(s_task)
        assert source.alias is not None
        assert source.alias == Tortoise.get_connection("default")

        db_data: Optional[PeriodicTask] = await PeriodicTask.filter(
            schedule_id=s_task.schedule_id
        ).first()
        assert db_data is not None
        assert origin_db_data.total_run_count + 1 == db_data.total_run_count


class TestTortoiseScheduleSourceErrors(object):
    """Test error conditions in TortoiseScheduleSource."""

    def test_tortoise_schedule_source_init_no_database_settings(self) -> None:
        """Test TortoiseScheduleSource initialization with no DATABASE settings."""
        from unittest.mock import MagicMock, patch

        import pytest

        # Mock settings to have no DATABASE configuration
        mock_unfazed_settings = MagicMock()
        mock_unfazed_settings.DATABASE = None

        with patch(
            "unfazed_taskiq.contrib.scheduler.sources.settings"
        ) as mock_settings_obj:
            mock_settings_obj.__getitem__.return_value = mock_unfazed_settings

            with pytest.raises(
                RuntimeError, match="DATABASE is not set in UNFAZED_SETTINGS"
            ):
                TortoiseScheduleSource()

    def test_tortoise_schedule_source_init_invalid_db_alias(self) -> None:
        """Test TortoiseScheduleSource initialization with invalid db_alias."""
        from unittest.mock import MagicMock, patch

        import pytest

        # Mock settings with DATABASE but without the requested alias
        mock_database = MagicMock()
        mock_database.connections = {"default": "some_connection"}  # Only has 'default'

        mock_unfazed_settings = MagicMock()
        mock_unfazed_settings.DATABASE = mock_database

        with patch(
            "unfazed_taskiq.contrib.scheduler.sources.settings"
        ) as mock_settings_obj:
            mock_settings_obj.__getitem__.return_value = mock_unfazed_settings

            with pytest.raises(
                RuntimeError, match="Database connection invalid_alias not found"
            ):
                TortoiseScheduleSource(db_alias="invalid_alias")

    async def test_tortoise_schedule_source_startup_tortoise_not_initialized(
        self,
    ) -> None:
        """Test TortoiseScheduleSource startup when Tortoise is not initialized."""
        from unittest.mock import patch

        import pytest

        source = TortoiseScheduleSource(schedule_alias="test")

        # Mock Tortoise._inited to be False
        with patch(
            "unfazed_taskiq.contrib.scheduler.sources.Tortoise"
        ) as mock_tortoise:
            mock_tortoise._inited = False

            with pytest.raises(RuntimeError, match="Tortoise is not initialized"):
                await source.startup()

    async def test_tortoise_schedule_source_add_schedule_duplicate_error(self) -> None:
        """Test TortoiseScheduleSource add_schedule with duplicate schedule_id."""
        import uuid

        import pytest

        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()

        # First create a task to ensure we have something to duplicate
        original_schedule_id = str(uuid.uuid4().hex)
        original_task = ScheduledTask(
            task_name="test.original_task",
            args=[],
            kwargs={},
            labels={},
            schedule_id=original_schedule_id,
            cron="0 0 * * *",
        )
        await source.add_schedule(original_task)

        # Now try to create a duplicate with the same schedule_id
        duplicate_task = ScheduledTask(
            task_name="test.duplicate_task",
            args=[],
            kwargs={},
            labels={},
            schedule_id=original_schedule_id,  # Use same schedule_id
            cron="0 1 * * *",
        )

        with pytest.raises(
            RuntimeError, match=f"Schedule {original_schedule_id} already exists"
        ):
            await source.add_schedule(duplicate_task)

    async def test_tortoise_schedule_source_add_schedule_with_time(self) -> None:
        """Test TortoiseScheduleSource add_schedule with time-based scheduling."""
        import uuid
        from datetime import datetime, timezone

        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()

        # Create a schedule task with time instead of cron
        task_time = datetime(2024, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        time_task = ScheduledTask(
            task_name="test.time_based_task",
            args=[1, 2, 3],
            kwargs={"key": "value"},
            labels={"env": "test"},
            schedule_id=str(uuid.uuid4().hex),
            time=task_time,  # Use time instead of cron
        )

        await source.add_schedule(time_task)

        # Verify the task was added correctly
        added_task = await PeriodicTask.filter(
            schedule_id=time_task.schedule_id
        ).first()
        assert added_task is not None
        assert added_task.task_name == "test.time_based_task"
        assert added_task.time == task_time
        assert added_task.cron is None

    async def test_tortoise_schedule_source_add_schedule_no_schedule_error(
        self,
    ) -> None:
        """Test TortoiseScheduleSource add_schedule with no cron or time."""
        import uuid
        from unittest.mock import MagicMock

        import pytest

        source = TortoiseScheduleSource(schedule_alias="test_schedule3")
        await source.startup()

        # Create a mock ScheduledTask that bypasses Pydantic validation
        # to test the RuntimeError in add_schedule method
        mock_task = MagicMock()
        mock_task.task_name = "test.no_schedule_task"
        mock_task.args = []
        mock_task.kwargs = {}
        mock_task.labels = {}
        mock_task.schedule_id = str(uuid.uuid4().hex)
        mock_task.cron = None
        mock_task.time = None

        with pytest.raises(RuntimeError, match="No schedule found"):
            await source.add_schedule(mock_task)



import asyncio
from typing import Any, Optional
from unittest.mock import AsyncMock, Mock, patch

from unfazed_taskiq.cli.worker.cmd import WorkerCMD


class TestWorkerCMD(object):
    def test_init_unfazed(self) -> None:
        """Test init_unfazed method."""
        cmd = WorkerCMD()

        with patch("unfazed_taskiq.cli.worker.cmd.Unfazed") as mock_unfazed:
            mock_unfazed_instance = Mock()
            mock_unfazed_instance.setup = AsyncMock(return_value=None)
            mock_unfazed.return_value = mock_unfazed_instance

            # Test the async method using asyncio.run
            # Create a new event loop to avoid conflicts
            asyncio.run(cmd.init_unfazed())

            # Verify Unfazed was instantiated
            mock_unfazed.assert_called_once()
            # Verify setup was called
            mock_unfazed_instance.setup.assert_called_once()
            # Verify unfazed attribute was set
            assert cmd.unfazed == mock_unfazed_instance

    def test_exec_success(self) -> None:
        """Test exec method with successful execution."""
        cmd = WorkerCMD()
        args = ["test_broker", "test_module"]

        with (
            patch("unfazed_taskiq.cli.worker.cmd.WorkerArgs") as mock_worker_args,
            patch("unfazed_taskiq.cli.worker.cmd.asyncio.run") as mock_asyncio_run,
            patch("unfazed_taskiq.cli.worker.cmd.run_worker") as mock_run_worker,
        ):
            # Setup mocks
            mock_worker_args_instance = Mock()
            mock_worker_args.from_cli.return_value = mock_worker_args_instance
            mock_run_worker.return_value = 0

            # Mock asyncio.run to properly handle coroutines
            def mock_asyncio_run_func(coro: Any) -> Any:
                if asyncio.iscoroutine(coro):
                    # Create a new event loop and run the coroutine
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    try:
                        return loop.run_until_complete(coro)
                    finally:
                        loop.close()
                return coro

            mock_asyncio_run.side_effect = mock_asyncio_run_func

            # Call exec
            result: Optional[int] = cmd.exec(args)

            # Verify calls
            mock_asyncio_run.assert_called_once()
            mock_worker_args.from_cli.assert_called_once_with(args)
            mock_run_worker.assert_called_once_with(mock_worker_args_instance)

            # Verify return value
            assert result == 0



import pytest

from unfazed_taskiq.registry.task import RegistryTask, RegistryTaskSchema


def sample_task(x: int, y: str = "hi") -> None:
    """Sample task used for registry tests."""


def auxiliary_task(flag: bool) -> None:
    pass


class TestRegistryTask:
    def setup_method(self) -> None:
        self.registry = RegistryTask()

    def test_register_broker_creates_schema(self) -> None:
        schedule = [{"cron": "* * * * *"}]
        self.registry.register_broker(
            sample_task, alias_name="alpha", schedule=schedule
        )
        path = f"{sample_task.__module__}.{sample_task.__name__}"
        schema = self.registry.get(path)
        assert isinstance(schema, RegistryTaskSchema)
        assert schema.name == "sample_task"
        assert schema.alias_name == "alpha"
        assert schema.docs == "Sample task used for registry tests."
        assert schema.schedule == schedule
        assert len(schema.params) == 2
        first, second = schema.params
        assert (
            first.name == "x"
            and first.required
            and first.default is None
            and first.hint_type is int
        )
        assert (
            second.name == "y"
            and not second.required
            and second.default == "hi"
            and second.hint_type is str
        )

    def test_duplicate_registration_raises(self) -> None:
        self.registry.register_broker(sample_task)
        with pytest.raises(ValueError):
            self.registry.register_broker(sample_task)

    def test_get_and_filter_path(self) -> None:
        self.registry.register_broker(sample_task, alias_name="alpha")
        self.registry.register_broker(auxiliary_task, alias_name="beta")
        assert self.registry.get("missing") is None
        assert len(self.registry.filter_path()) == 2
        filtered = self.registry.filter_path("auxiliary_task")
        assert len(filtered) == 1 and filtered[0].name == "auxiliary_task"
        assert self.registry.filter_path("absent") == []



import typing as t

from pydantic import BaseModel, ConfigDict, Field
from taskiq import ScheduleSource
from taskiq.events import TaskiqEvents
from unfazed.conf import register_settings


class Broker(BaseModel):
    backend: str = Field(alias="BACKEND")
    options: t.Optional[t.Dict[str, t.Any]] = Field(default=None, alias="OPTIONS")
    middlewares: t.List[str] = Field(
        default=[], alias="MIDDLEWARES"
    )  # unfazed_taskiq.middleware.UnfazedTaskiqExceptionMiddleware
    handlers: t.List[t.Dict[str, t.Union[str, TaskiqEvents]]] = Field(
        default=[], alias="HANDLERS"
    )


class Result(BaseModel):
    backend: str = Field(alias="BACKEND")
    options: t.Optional[t.Dict[str, t.Any]] = Field(default=None, alias="OPTIONS")


class Scheduler(BaseModel):
    backend: str = Field(alias="BACKEND")
    sources: t.Optional[t.List[t.Union[str, ScheduleSource]]] = Field(
        default=None, alias="SOURCES"
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)


class TaskiqConfig(BaseModel):
    broker: Broker = Field(alias="BROKER")
    result: t.Optional[Result] = Field(default=None, alias="RESULT")
    scheduler: t.Optional[Scheduler] = Field(default=None, alias="SCHEDULER")


@register_settings("UNFAZED_TASKIQ_SETTINGS")
class UnfazedTaskiqSettings(BaseModel):
    taskiq_config: t.Dict[str, TaskiqConfig] = Field(alias="TASKIQ_CONFIG")
    default_alias_name: t.Optional[str] = Field(
        alias="DEFAULT_ALIAS_NAME", default="default"
    )



import os

from unfazed_taskiq.contrib.scheduler.sources import TortoiseScheduleSource

REDIS_HOST = os.getenv("REDIS_HOST", "redis")

UNFAZED_SETTINGS = {
    "LIFESPAN": [],
    "ROOT_URLCONF": "tests.proj.entry.routes",
    "INSTALLED_APPS": ["tests.proj.app1", "unfazed_taskiq.contrib.scheduler"],
    "DATABASE": {
        "CONNECTIONS": {
            "default": {
                "ENGINE": "tortoise.backends.mysql",
                "CREDENTIALS": {
                    "HOST": os.getenv("MYSQL_HOST", "mysql"),
                    "PORT": os.getenv("MYSQL_PORT", 3306),
                    "USER": os.getenv("MYSQL_USER", "root"),
                    "PASSWORD": os.getenv("MYSQL_PASSWORD", "app"),
                    "DATABASE": "test_app",
                },
            },
        },
    },
    # "LOGGING": {
    #     "formatters": {
    #         "taskiq_format": {
    #             "format": "[%(asctime)s][%(levelname)-7s] [%(module)s:%(funcName)s:%(lineno)d] %(message)s"
    #         },
    #     },
    #     "handlers": {
    #         "_console": {
    #             "class": "logging.StreamHandler",
    #             "level": "DEBUG",
    #             "formatter": "taskiq_format",
    #         }
    #     },
    #     "loggers": {
    #         "taskiq": {
    #             "handlers": ["_console"],
    #             "level": "INFO",
    #         },
    #     },
    # },
}


DEFAULT_AMQP_URL = "amqp://app:app@rabbitmq:5672"
AMQP_URL = os.getenv("AMQP_URL", DEFAULT_AMQP_URL)


default_taskiq_name = "testtaskiq"
source = TortoiseScheduleSource(schedule_alias=default_taskiq_name)

UNFAZED_TASKIQ_SETTINGS = {
    "DEFAULT_TASKIQ_NAME": default_taskiq_name,
    "TASKIQ_CONFIG": {
        default_taskiq_name: {
            "BROKER": {
                "BACKEND": "taskiq.InMemoryBroker",
                "OPTIONS": {},
            },
            "SCHEDULER": {
                "SOURCES": [source],
                "BACKEND": "taskiq.TaskiqScheduler",
            },
        },
    },
}



from typing import List, Optional

from pydantic import BaseModel, ConfigDict
from taskiq import AsyncBroker, ScheduleSource, TaskiqEvents, TaskiqScheduler
from unfazed.utils import import_string

from unfazed_taskiq.settings import TaskiqConfig


class TaskiqAgent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    alias_name: str
    broker: AsyncBroker
    scheduler: Optional[TaskiqScheduler]
    config: TaskiqConfig

    @classmethod
    def setup(cls, alias_name: str, config: TaskiqConfig) -> "TaskiqAgent":
        # setup broker
        broker_cls = import_string(config.broker.backend)
        broker_options = config.broker.options or {}
        broker: AsyncBroker = broker_cls(**broker_options)

        # setup middlewares
        for middleware_path in config.broker.middlewares:
            if middleware_path:  # Skip empty middleware paths
                middleware_cls = import_string(middleware_path)
                middleware = middleware_cls()
                broker.add_middlewares(middleware)

        # setup handlers
        for handler in config.broker.handlers:
            handler_cls = import_string(handler["handler"])
            event = handler["event"]
            if isinstance(event, str):
                event = TaskiqEvents(event)
            broker.add_event_handler(event, handler_cls)

        # setup result backend
        if config.result:
            result_cls = import_string(config.result.backend)
            result_options = config.result.options or {}
            broker.with_result_backend(result_cls(**result_options))

        # setup scheduler
        scheduler = None
        if config.scheduler:
            scheduler_cls: type[TaskiqScheduler] = import_string(
                config.scheduler.backend
            )
            scheduler_sources = config.scheduler.sources or []
            sources: List[ScheduleSource] = []
            for source in scheduler_sources:
                if isinstance(source, ScheduleSource):
                    sources.append(source)
                else:
                    source_cls = import_string(source)
                    sources.append(source_cls(broker))
            scheduler = scheduler_cls(broker=broker, sources=sources)

        return cls(
            alias_name=alias_name, broker=broker, scheduler=scheduler, config=config
        )

    async def startup(self) -> None:
        if self.scheduler and isinstance(self.scheduler, TaskiqScheduler):
            await self.scheduler.startup()
        await self.broker.startup()

    async def shutdown(self) -> None:
        if self.scheduler and isinstance(self.scheduler, TaskiqScheduler):
            await self.scheduler.shutdown()
        await self.broker.shutdown()



from argparse import ZERO_OR_MORE, ArgumentDefaultsHelpFormatter, ArgumentParser
from dataclasses import dataclass
from typing import List, Optional, Sequence, Union

from taskiq.cli.common_args import LogLevel
from taskiq.cli.scheduler.args import SchedulerArgs
from taskiq.scheduler.scheduler import TaskiqScheduler


@dataclass
class SchedulerEventArgs(SchedulerArgs):
    """Arguments for scheduler."""

    scheduler: Union[str, TaskiqScheduler]
    modules: List[str]
    log_level: str = LogLevel.INFO.name
    configure_logging: bool = True
    fs_discover: bool = False
    tasks_pattern: Sequence[str] = ("**/tasks.py",)
    skip_first_run: bool = False
    update_interval: Optional[int] = None
    alias_name: Sequence[str] = ()

    @classmethod
    def from_cli(cls, args: Optional[Sequence[str]] = None) -> "SchedulerEventArgs":
        """
        Build scheduler args from CLI arguments.

        This method takes arguments as args parameter.

        :param args: current CLI arguments, defaults to None
        :return: instance of scheduler args.
        """
        parser = ArgumentParser(
            formatter_class=ArgumentDefaultsHelpFormatter,
            description="Subcommand to run scheduler",
        )
        parser.add_argument(
            "scheduler",
            help="Path to scheduler or scheduler factory function",
        )
        parser.add_argument(
            "modules",
            help="List of modules where to look for tasks.",
            nargs=ZERO_OR_MORE,
        )
        parser.add_argument(
            "--fs-discover",
            "-fsd",
            action="store_true",
            help=(
                "If this option is on, "
                "taskiq will try to find tasks modules "
                "in current directory recursively. Name of file to search for "
                "can be configured using `--tasks-pattern` option."
            ),
        )
        parser.add_argument(
            "--tasks-pattern",
            "-tp",
            default=["**/tasks.py"],
            action="append",
            help="Glob patterns of files in which taskiq will try to find the tasks.",
        )
        parser.add_argument(
            "--log-level",
            default=LogLevel.INFO.name,
            choices=[level.name for level in LogLevel],
            help="scheduler log level",
        )
        parser.add_argument(
            "--no-configure-logging",
            action="store_false",
            dest="configure_logging",
            help="Use this parameter if your application configures custom logging.",
        )
        parser.add_argument(
            "--skip-first-run",
            action="store_true",
            dest="skip_first_run",
            help=(
                "Skip first run of scheduler. "
                "This option skips running tasks immediately after scheduler start."
            ),
        )
        parser.add_argument(
            "--update-interval",
            type=int,
            default=None,
            help=(
                "Interval in seconds to check for new tasks. "
                "If not specified, scheduler will run once a minute."
            ),
        )
        parser.add_argument(
            "--alias-name",
            "-an",
            default=[],
            action="append",
            help="should run the scheduler with the given alias",
        )

        namespace = parser.parse_args(args)
        # If there are any patterns specified, remove default.
        # This is an argparse limitation.
        if len(namespace.tasks_pattern) > 1:
            namespace.tasks_pattern.pop(0)
        return cls(**namespace.__dict__)



import pytest
from unfazed.core import Unfazed
from unfazed.test import Requestfactory


@pytest.mark.asyncio
async def test_api(unfazed: Unfazed) -> None:
    async with Requestfactory(unfazed) as rf:
        resp = await rf.get("/app1/add?a=1&b=2")

        assert resp.status_code == 200
        assert resp.json() == {"result": 3}



"""
Unit tests for middleware module.

This module contains tests for the UnfazedTaskiqExceptionMiddleware functionality,
covering error handling, logging, and Sentry integration.
"""

from unittest.mock import MagicMock, patch

import pytest
from taskiq import TaskiqMessage, TaskiqResult

from unfazed_taskiq.middleware import UnfazedTaskiqExceptionMiddleware


class TestUnfazedTaskiqExceptionMiddleware:
    """Test UnfazedTaskiqExceptionMiddleware functionality."""

    def setup_method(self) -> None:
        """Set up test fixtures before each test method."""
        self.middleware = UnfazedTaskiqExceptionMiddleware()
        self.mock_message = MagicMock(spec=TaskiqMessage)
        self.mock_message.task_name = "test_task"
        self.mock_message.args = ("arg1", "arg2")
        self.mock_message.kwargs = {"key1": "value1", "key2": "value2"}

        self.mock_result = MagicMock(spec=TaskiqResult)
        self.test_exception = ValueError("Test error message")

    async def test_on_error_calls_capture_exception(self) -> None:
        """Test that on_error calls capture_exception with correct parameters."""
        with patch("unfazed_taskiq.middleware.capture_exception") as mock_capture:
            await self.middleware.on_error(
                self.mock_message, self.mock_result, self.test_exception
            )

            mock_capture.assert_called_once_with(
                self.test_exception, result=self.mock_result, message=self.mock_message
            )

    @pytest.mark.asyncio
    async def test_on_error_logs_error_with_correct_format(self) -> None:
        """Test that on_error logs error with correct format and extra data."""
        with patch("unfazed_taskiq.middleware.capture_exception"):
            with patch("unfazed_taskiq.middleware.log") as mock_logger:
                await self.middleware.on_error(
                    self.mock_message, self.mock_result, self.test_exception
                )

                # Verify error log was called
                mock_logger.error.assert_called_once()

                # Get the logged message and extra data
                call_args = mock_logger.error.call_args
                logged_message = call_args[0][0]
                extra_data = call_args[1]["extra"]

                # Verify message format
                expected_message = (
                    f"Task '{self.mock_message.task_name}' failed with "
                    f"{type(self.test_exception).__name__}: {self.test_exception}"
                )
                assert logged_message == expected_message

                # Verify extra data contains all required fields
                assert extra_data["task_name"] == self.mock_message.task_name
                assert extra_data["task_args"] == self.mock_message.args
                assert extra_data["task_kwargs"] == self.mock_message.kwargs
                assert (
                    extra_data["exception_type"] == type(self.test_exception).__name__
                )
                assert extra_data["exception"] == str(self.test_exception)
                assert "traceback" in extra_data



import traceback
from typing import Any

from taskiq import TaskiqMessage, TaskiqResult
from taskiq.abc.middleware import TaskiqMiddleware
from unfazed_sentry import capture_exception

from unfazed_taskiq.logger import log


class UnfazedTaskiqExceptionMiddleware(TaskiqMiddleware):
    async def on_error(
        self,
        message: "TaskiqMessage",
        result: "TaskiqResult[Any]",
        exception: BaseException,
    ) -> None:
        capture_exception(exception, result=result, message=message)
        exception_type = type(exception).__name__

        log.error(
            f"Task '{message.task_name}' failed with {exception_type}: {exception}",
            extra={
                "task_name": message.task_name,
                "task_args": message.args,
                "task_kwargs": message.kwargs,
                "exception_type": exception_type,
                "exception": str(exception),
                "traceback": traceback.format_exc(),
            },
        )



from typing import Any

from taskiq import TaskiqResult
from unfazed.http import HttpRequest, JsonResponse

from .tasks import add


async def add_endpoint(request: HttpRequest, a: int, b: int) -> JsonResponse:
    task: Any = await add.kiq(a, b)  # type: ignore[attr-defined]

    result: TaskiqResult = await task.wait_result(timeout=10)

    return JsonResponse({"result": result.return_value})



"""
Unit tests for TaskiqLifeSpan class.

This module contains tests for the TaskiqLifeSpan class functionality,
covering initialization, startup, and shutdown operations.
"""

from unittest.mock import AsyncMock, MagicMock, patch

from unfazed_taskiq.lifespan import TaskiqLifeSpan


class TestTaskiqLifeSpan:
    """Test TaskiqLifeSpan functionality."""

    def test_taskiq_lifespan_init(self) -> None:
        """Test TaskiqLifeSpan initialization."""
        mock_unfazed = MagicMock()

        with patch("unfazed_taskiq.lifespan.agents") as agent_mock:
            agent_mock.setup.return_value = None

            lifespan = TaskiqLifeSpan(mock_unfazed)

            assert lifespan.unfazed is mock_unfazed
            assert lifespan.agents is agent_mock

            agent_mock.setup.assert_called_once_with()

    async def test_taskiq_lifespan_on_startup(self) -> None:
        """Test TaskiqLifeSpan on_startup method."""
        mock_unfazed = MagicMock()

        with patch("unfazed_taskiq.lifespan.agents") as agent_mock:
            agent_mock.setup.return_value = None
            agent_mock.startup = AsyncMock()

            lifespan = TaskiqLifeSpan(mock_unfazed)

            await lifespan.on_startup()

            agent_mock.startup.assert_awaited_once_with()

    async def test_taskiq_lifespan_on_shutdown(self) -> None:
        """Test TaskiqLifeSpan on_shutdown method."""
        mock_unfazed = MagicMock()

        with patch("unfazed_taskiq.lifespan.agents") as agent_mock:
            agent_mock.setup.return_value = None
            agent_mock.shutdown = AsyncMock()

            lifespan = TaskiqLifeSpan(mock_unfazed)

            await lifespan.on_shutdown()

            agent_mock.shutdown.assert_awaited_once_with()



import os
from typing import Optional

from taskiq import AsyncBroker, TaskiqScheduler
from unfazed.utils import Storage, import_setting

from unfazed_taskiq.agent.model import TaskiqAgent
from unfazed_taskiq.settings import UnfazedTaskiqSettings


class AgentHandler(Storage[TaskiqAgent]):
    def __init__(self) -> None:
        super().__init__()
        self.default_alias_name: str = "default"  # Default fallback
        self._ready = False
        self.check_ready()

    def register(self, alias_name: str, agent: TaskiqAgent) -> None:
        if alias_name in self.storage:
            raise ValueError(f"Agent {alias_name} already registered")
        self.storage[alias_name] = agent

    def reset(self) -> None:
        self.clear()
        self._ready = False

    def setup(self) -> None:
        if not os.environ.get("UNFAZED_SETTINGS_MODULE"):
            raise ValueError("UNFAZED_SETTINGS_MODULE is not set")

        try:
            # extract settings from unfazed settings
            settings_kv = import_setting("UNFAZED_SETTINGS_MODULE")
        except ImportError as e:
            raise ImportError(f"Failed to import settings module: {e}")

        # extract broker settings
        taskiq_settings = settings_kv["UNFAZED_TASKIQ_SETTINGS"]
        self.default_alias_name = taskiq_settings.get("DEFAULT_TASKIQ_NAME", "default")
        try:
            taskiq_config_settings = UnfazedTaskiqSettings.model_validate(
                taskiq_settings
            )
        except Exception as e:
            raise ValueError(f"Invalid settings configuration: {e}")

        for alias_name, taskiq_config in taskiq_config_settings.taskiq_config.items():
            taskiq_agent: TaskiqAgent = TaskiqAgent.setup(alias_name, taskiq_config)
            self.register(alias_name, taskiq_agent)
        self._ready = True

    def check_ready(self) -> None:
        if not self._ready:
            self.reset()
            self.setup()

    def get_agent(self, alias_name: Optional[str]) -> Optional[TaskiqAgent]:
        """Get the agent by alias name"""
        if not self._ready:
            self.check_ready()
        _alias_name = self.default_alias_name if alias_name is None else alias_name
        return self.storage.get(_alias_name, None)

    @property
    def scheduler(self) -> TaskiqScheduler:
        """Get the default scheduler"""
        if not self._ready:
            self.check_ready()
        return self.storage[self.default_alias_name].scheduler

    @property
    def broker(self) -> AsyncBroker:
        """Get the default broker"""
        if not self._ready:
            self.check_ready()
        return self.storage[self.default_alias_name].broker

    async def startup(self) -> None:
        for agent_model in self.storage.values():
            await agent_model.startup()

    async def shutdown(self) -> None:
        for agent_model in self.storage.values():
            await agent_model.shutdown()


agents = AgentHandler()
scheduler = agents.scheduler
broker = agents.broker



from unfazed.contrib.admin.registry import ModelAdmin, register

from . import serializer as s


@register(s.PeriodicTaskSerializer)
class PeriodicTaskAdmin(ModelAdmin):
    route_label: str = "TaskIQ"
    component: str = "ModelAdmin"

    list_display: list[str] = [
        "schedule_alias",
        "name",
        "description",
        "task_name",
        "task_args",
        "task_kwargs",
        "labels",
        "cron",
        "last_run_at",
        "time",
        "total_run_count",
        "enabled",
    ]
    search_fields: list[str] = [
        "schedule_alias",
        "name",
        "task_name",
    ]
    list_search: list[str] = [
        "schedule_alias",
        "name",
        "task_name",
    ]
    detail_display: list[str] = [
        "schedule_alias",
        "name",
        "description",
        "task_name",
        "task_args",
        "task_kwargs",
        "labels",
        "cron",
        "time",
        "last_run_at",
        "total_run_count",
        "enabled",
        "schedule_id",
    ]
    readonly_fields: list[str] = [
        "last_run_at",
        "total_run_count",
        "schedule_id",
    ]



from typing import Any, Callable, Optional

from unfazed_taskiq.agent.handler import agents
from unfazed_taskiq.agent.model import TaskiqAgent
from unfazed_taskiq.registry.task import rs


def task(
    func: Optional[Callable] = None,
    *,
    alias_name: Optional[str] = None,
    **task_kwargs: Any,
) -> Callable:
    """
    use taskiq decorator to decorate task

    Args:
        func: function to decorate (when used without parentheses)
        alias_name: alias name, if None, use default alias
        **task_kwargs: other arguments for taskiq task decorator

    Example:
        @task
        async def simple_task():
            pass

        @task(alias_name="high_priority")
        async def high_priority_task():
            pass

        @task(alias_name="low_priority", schedule=[{"cron": "*/5 * * * *"}])
        async def scheduled_task():
            pass
    """

    def decorator(func: Callable) -> Callable:
        _agent: Optional[TaskiqAgent] = agents.get_agent(alias_name)
        rs.register_broker(func, alias_name, **task_kwargs)
        # decorate task
        if _agent is None:
            raise ValueError(f"Agent {alias_name} not found")
        return _agent.broker.task(**task_kwargs)(func)

    # Support @task and @task()
    return decorator if func is None else decorator(func)



from dataclasses import replace
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from taskiq.cli.common_args import LogLevel

from unfazed_taskiq.cli.scheduler.args import SchedulerEventArgs
from unfazed_taskiq.cli.scheduler.cmd import SchedulerCMD


class TestSchedulerCMD:
    def test_short_help(self) -> None:
        assert SchedulerCMD().short_help == "Run task scheduler"

    @pytest.mark.asyncio
    async def test_init_unfazed(self) -> None:
        cmd = SchedulerCMD()
        unfazed = AsyncMock()
        with patch("unfazed_taskiq.cli.scheduler.cmd.Unfazed", return_value=unfazed):
            await cmd.init_unfazed()
        unfazed.setup.assert_awaited_once()
        assert cmd.unfazed is unfazed

    def _run_exec(
        self, alias_name: list[str] | None, storage: dict[str, Any]
    ) -> MagicMock:
        cmd = SchedulerCMD()
        args = ["module", "pkg"] + (
            [] if alias_name is None else ["--scheduler-alias", *alias_name]
        )
        parsed = SchedulerEventArgs(
            scheduler="scheduler.path",
            modules=["pkg"],
            alias_name=[] if alias_name is None else alias_name,
        )
        run_spy = AsyncMock()
        agents = {alias: MagicMock(scheduler=sched) for alias, sched in storage.items()}
        with (
            patch("unfazed_taskiq.cli.scheduler.cmd.Unfazed", return_value=AsyncMock()),
            patch(
                "unfazed_taskiq.cli.scheduler.cmd.SchedulerEventArgs.from_cli",
                return_value=parsed,
            ),
            patch("unfazed_taskiq.cli.scheduler.cmd.agents.storage", agents),
            patch("unfazed_taskiq.cli.scheduler.cmd.run_scheduler", run_spy),
            patch("asyncio.create_task", side_effect=lambda coro: coro),
        ):
            cmd.exec(args)
        return run_spy

    def test_exec_with_alias(self) -> None:
        run_spy = self._run_exec(["alpha"], {"alpha": AsyncMock(), "beta": AsyncMock()})
        assert run_spy.await_count == 1

    def test_exec_without_alias(self) -> None:
        run_spy = self._run_exec(None, {"alpha": AsyncMock(), "beta": AsyncMock()})
        assert run_spy.await_count == 2


class TestSchedulerEventArgs:
    def test_defaults(self) -> None:
        args = SchedulerEventArgs(scheduler="path", modules=["pkg"])
        assert args.log_level == LogLevel.INFO.name
        assert args.tasks_pattern == ("**/tasks.py",)

    def test_override(self) -> None:
        args = SchedulerEventArgs(
            scheduler="path",
            modules=["p1", "p2"],
            log_level=LogLevel.DEBUG.name,
            configure_logging=False,
            fs_discover=True,
            tasks_pattern=("**/tasks.py", "**/jobs.py"),
            skip_first_run=True,
            update_interval=10,
            alias_name=("alpha",),
        )
        assert not args.configure_logging
        assert args.alias_name == ("alpha",)

    def test_from_cli_parsing(self) -> None:
        args = SchedulerEventArgs.from_cli(
            [
                "scheduler",
                "module",
                "--fs-discover",
                "--tasks-pattern",
                "**/jobs.py",
                "--log-level",
                "DEBUG",
                "--no-configure-logging",
                "--skip-first-run",
                "--update-interval",
                "30",
                "--alias-name",
                "alpha",
            ]
        )
        assert args.modules == ["module"]
        assert args.tasks_pattern == ["**/jobs.py"]
        assert args.alias_name == ["alpha"]
        assert args.update_interval == 30

    def test_replace(self) -> None:
        original = SchedulerEventArgs(scheduler="a", modules=["m"])
        updated = replace(original, log_level=LogLevel.WARNING.name, alias_name=("x",))
        assert updated.log_level == LogLevel.WARNING.name
        assert original.log_level == LogLevel.INFO.name
        assert updated.alias_name == ("x",)



import asyncio
from typing import Optional, Sequence

from taskiq.abc.cmd import TaskiqCMD
from taskiq.cli.worker.args import WorkerArgs
from taskiq.cli.worker.run import run_worker
from unfazed.core import Unfazed


class WorkerCMD(TaskiqCMD):
    """Command to run workers."""

    short_help = "Helper to run workers"

    async def init_unfazed(self) -> None:
        self.unfazed = Unfazed(silent=True)
        await self.unfazed.setup()

    def exec(self, args: Sequence[str]) -> Optional[int]:
        """
        Start worker process.

        Worker process creates several small
        processes in which tasks are actually processed.

        :param args: CLI arguments.
        :returns: status code.
        """
        # setup unfazed
        asyncio.run(self.init_unfazed())
        # setup worker
        wargs: WorkerArgs = WorkerArgs.from_cli(args)
        return run_worker(wargs)



from types import SimpleNamespace
from typing import Any, Callable

import pytest

from unfazed_taskiq import decorators


class DummyBroker:
    def __init__(self) -> None:
        self.calls: list[dict[str, Any]] = []

    def task(self, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        self.calls.append(kwargs)

        def _decorator(func: Any) -> Callable[..., Any]:
            def _wrapper(*args: Any, **inner_kwargs: Any) -> Any:
                return func(*args, **inner_kwargs)

            return _wrapper

        return _decorator


def test_task_decorator_with_parameters(monkeypatch: pytest.MonkeyPatch) -> None:
    broker = DummyBroker()
    agent = SimpleNamespace(broker=broker)
    captured_aliases: list = []
    register_calls: list = []

    def fake_get_agent(alias_name: str) -> Any:
        captured_aliases.append(alias_name)
        return agent

    def fake_register(func: Any, alias_name: str, **kwargs: Any) -> None:
        register_calls.append((func, alias_name, kwargs))

    monkeypatch.setattr(decorators.agents, "get_agent", fake_get_agent)
    monkeypatch.setattr(decorators.rs, "register_broker", fake_register)

    @decorators.task(alias_name="alpha", schedule=[{"cron": "* * * * *"}])
    def sample_task(value: int) -> int:
        return value + 1

    assert sample_task(2) == 3
    assert captured_aliases == ["alpha"]
    assert broker.calls == [{"schedule": [{"cron": "* * * * *"}]}]
    assert register_calls[0][0].__name__ == "sample_task"
    assert register_calls[0][1] == "alpha"
    assert register_calls[0][2] == {"schedule": [{"cron": "* * * * *"}]}


def test_task_decorator_without_parameters(monkeypatch: pytest.MonkeyPatch) -> None:
    broker = DummyBroker()
    agent = SimpleNamespace(broker=broker)
    register_calls = []

    monkeypatch.setattr(decorators.agents, "get_agent", lambda alias_name: agent)
    monkeypatch.setattr(
        decorators.rs,
        "register_broker",
        lambda func, alias_name, **kwargs: register_calls.append((alias_name, kwargs)),
    )

    @decorators.task
    def bare_task() -> str:
        return "ok"

    assert bare_task() == "ok"
    assert register_calls == [(None, {})]
    assert broker.calls == [{}]


def test_task_decorator_missing_agent(monkeypatch: pytest.MonkeyPatch) -> None:
    register_calls = []

    def fake_register(_: Any, alias_name: str, **kwargs: Any) -> None:
        register_calls.append((alias_name, kwargs))

    monkeypatch.setattr(decorators.agents, "get_agent", lambda alias_name: None)
    monkeypatch.setattr(decorators.rs, "register_broker", fake_register)

    def ghost_task() -> str:
        return "ghost"

    with pytest.raises(ValueError, match="Agent ghost not found"):
        decorators.task(alias_name="ghost")(ghost_task)

    assert register_calls == [("ghost", {})]



import sys
from typing import Any, Awaitable, Callable, Iterable, List, Sequence
from unittest.mock import AsyncMock

import pytest
from taskiq import (
    AckableMessage,
    AsyncBroker,
    ScheduleSource,
    TaskiqEvents,
    TaskiqScheduler,
)
from taskiq.abc.middleware import TaskiqMiddleware
from taskiq.abc.result_backend import AsyncResultBackend
from taskiq.message import BrokerMessage
from taskiq.result import TaskiqResult
from taskiq.state import TaskiqState

from unfazed_taskiq.agent.model import TaskiqAgent
from unfazed_taskiq.settings import Broker, Result, Scheduler, TaskiqConfig


class SchedulerResult:
    def __init__(self) -> None:
        self.tasks: list[str] = []


class TestTaskiqAgent:
    def _build_config(self) -> TaskiqConfig:
        return TaskiqConfig(
            BROKER=Broker(
                BACKEND="tests.doubles.FakeBroker",
                OPTIONS={"foo": "bar"},
                MIDDLEWARES=[
                    "tests.doubles.MiddlewareA",
                    "tests.doubles.MiddlewareB",
                    "",
                ],
                HANDLERS=[
                    {
                        "handler": "tests.doubles.EventHandler",
                        "event": TaskiqEvents.WORKER_STARTUP.value,
                    },
                    {
                        "handler": "tests.doubles.EventHandler",
                        "event": TaskiqEvents.CLIENT_STARTUP,
                    },
                ],
            ),
            RESULT=Result(BACKEND="tests.doubles.ResultBackend", OPTIONS={"ttl": 5}),
            SCHEDULER=Scheduler(
                BACKEND="tests.doubles.SchedulerBackend",
                SOURCES=[
                    "tests.doubles.SourceFactory",
                    "tests.doubles.BoundSource",
                ],
            ),
        )

    @pytest.fixture(autouse=True)
    def doubles(self, monkeypatch: pytest.MonkeyPatch) -> None:
        class FakeResultBackend(AsyncResultBackend[Any]):
            def __init__(self, **kwargs: Any) -> None:
                super().__init__()
                self.kwargs = kwargs
                self.calls: list[tuple[str, tuple[Any, ...], dict[str, Any]]] = []

            async def set_result(self, task_id: str, result: TaskiqResult[Any]) -> None:
                self.calls.append(("set", (task_id, result), {}))

            async def is_result_ready(self, task_id: str) -> bool:
                self.calls.append(("ready", (task_id,), {}))
                return True

            async def get_result(
                self, task_id: str, with_logs: bool = False
            ) -> TaskiqResult[Any]:
                self.calls.append(("get", (task_id,), {"with_logs": with_logs}))
                return TaskiqResult(  # type: ignore
                    is_ok=True, return_value=None, execution_time=0, log=None
                )

        class FakeBroker(AsyncBroker):
            def __init__(self, **kwargs: Any) -> None:
                super().__init__()
                self.kwargs = kwargs
                self.startup_mock: AsyncMock = AsyncMock()
                self.shutdown_mock: AsyncMock = AsyncMock()
                self.kick_mock: AsyncMock = AsyncMock()
                self.listen_result: Iterable[AckableMessage] = []

            async def kick(self, message: BrokerMessage) -> None:
                await self.kick_mock(message)

            async def listen(self) -> Iterable[AckableMessage]:  # type: ignore
                return self.listen_result

            async def startup(self) -> None:
                await self.startup_mock()
                await super().startup()

            async def shutdown(self) -> None:
                await self.shutdown_mock()
                await super().shutdown()

        class MiddlewareA(TaskiqMiddleware):
            async def pre_execute(self, message: BrokerMessage) -> BrokerMessage:  # type: ignore
                return message

        class MiddlewareB(TaskiqMiddleware):
            async def pre_execute(self, message: BrokerMessage) -> BrokerMessage:  # type: ignore
                return message

        class EventHandler:
            def __call__(self, state: TaskiqState) -> None:
                state.custom["called"] = True  # type: ignore[attr-defined]

        class SchedulerBackend(TaskiqScheduler):
            def __init__(
                self, broker: AsyncBroker, sources: Sequence[ScheduleSource]
            ) -> None:
                self.startup_mock: AsyncMock = AsyncMock()
                self.shutdown_mock: AsyncMock = AsyncMock()
                self.run_mock: AsyncMock = AsyncMock()
                super().__init__(broker, list(sources))

            async def add_task(self, task: Callable[..., Awaitable[None]]) -> None:
                self.run_mock(task)

            async def remove_task(self, task: Callable[..., Awaitable[None]]) -> None:
                self.run_mock(task)

            async def run(self) -> SchedulerResult:
                await self.run_mock()
                return SchedulerResult()

            async def startup(self) -> None:
                await self.startup_mock()

            async def shutdown(self) -> None:
                await self.shutdown_mock()

        class SourceFactory:
            def __init__(self, broker: AsyncBroker) -> None:
                self.broker = broker
                self.source = f"source-{id(broker)}"

        class BoundSource(ScheduleSource):
            def __init__(self, broker: AsyncBroker) -> None:
                self.broker = broker

            async def get_schedules(self) -> List[SchedulerResult]:  # type: ignore
                return []

        doubles_module = {
            "tests.doubles.FakeBroker": FakeBroker,
            "tests.doubles.MiddlewareA": MiddlewareA,
            "tests.doubles.MiddlewareB": MiddlewareB,
            "tests.doubles.EventHandler": EventHandler,
            "tests.doubles.ResultBackend": FakeResultBackend,
            "tests.doubles.SchedulerBackend": SchedulerBackend,
            "tests.doubles.SourceFactory": SourceFactory,
            "tests.doubles.BoundSource": BoundSource,
        }

        module = type(sys)("tests.doubles")
        for name, value in doubles_module.items():
            setattr(module, name.split(".")[-1], value)
        monkeypatch.setitem(sys.modules, "tests.doubles", module)

        def import_string_stub(path: str) -> Any:
            if path in doubles_module:
                return doubles_module[path]
            raise ImportError(path)

        monkeypatch.setattr(
            "unfazed_taskiq.agent.model.import_string", import_string_stub
        )

        self.fake_classes = {
            "broker": FakeBroker,
            "scheduler": SchedulerBackend,
            "result_backend": FakeResultBackend,
        }

    def test_setup_full_configuration(self) -> None:
        config = self._build_config()
        agent = TaskiqAgent.setup("alias", config)

        assert agent.alias_name == "alias"
        assert agent.config is config
        broker = agent.broker
        assert isinstance(broker, self.fake_classes["broker"])
        middleware_names = [
            middleware.__class__.__name__ for middleware in broker.middlewares
        ]
        assert middleware_names == ["MiddlewareA", "MiddlewareB"]
        assert broker.event_handlers[TaskiqEvents.WORKER_STARTUP]
        assert broker.event_handlers[TaskiqEvents.CLIENT_STARTUP]
        scheduler: TaskiqScheduler = agent.scheduler  # type: ignore
        assert isinstance(scheduler, self.fake_classes["scheduler"])
        assert scheduler.broker is broker
        assert len(scheduler.sources) == 2
        assert scheduler.sources[0].source.startswith("source-")  # type: ignore
        assert isinstance(scheduler.sources[1], ScheduleSource)

    def test_setup_without_optional_sections(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        config = TaskiqConfig(
            BROKER=Broker(BACKEND="tests.doubles.FakeBroker"),
            RESULT=None,
            SCHEDULER=None,
        )
        agent = TaskiqAgent.setup("alias", config)
        assert agent.scheduler is None
        assert isinstance(agent.broker, self.fake_classes["broker"])

    async def test_startup_and_shutdown(self) -> None:
        config = TaskiqConfig(
            BROKER=Broker(BACKEND="tests.doubles.FakeBroker"),
            RESULT=None,
            SCHEDULER=Scheduler(BACKEND="tests.doubles.SchedulerBackend", SOURCES=[]),
        )
        agent = TaskiqAgent.setup("alias", config)

        await agent.startup()
        assert agent.scheduler is not None
        agent.scheduler.startup_mock.assert_awaited_once()  # type: ignore
        agent.broker.startup_mock.assert_awaited_once()  # type: ignore

        await agent.shutdown()
        agent.scheduler.shutdown_mock.assert_awaited_once()  # type: ignore
        agent.broker.shutdown_mock.assert_awaited_once()  # type: ignore

    async def test_lifecycle_without_scheduler(self) -> None:
        config = TaskiqConfig(
            BROKER=Broker(BACKEND="tests.doubles.FakeBroker"),
            RESULT=None,
            SCHEDULER=None,
        )
        agent = TaskiqAgent.setup("alias", config)

        await agent.startup()
        agent.broker.startup_mock.assert_awaited_once()  # type: ignore

        await agent.shutdown()
        agent.broker.shutdown_mock.assert_awaited_once()  # type: ignore



import uuid

import orjson as json
from taskiq import ScheduledTask
from tortoise import fields, models


class BaseModel(models.Model):
    id = fields.IntField(pk=True)
    updated_at = fields.DatetimeField(auto_now=True)
    created_at = fields.DatetimeField(auto_now_add=True)

    class Meta:
        abstract = True


class PeriodicTask(BaseModel):
    class Meta:
        table = "unfazed_taskiq_periodic_task"

    schedule_alias = fields.CharField(
        max_length=255,
        description="The alias of the schedule.",
        default="default",
    )

    name = fields.CharField(
        max_length=255,
        default="NOT_GIVEN",
        description="The name of the task. eg: FF-SG-YOUTUBE",
    )

    description = fields.TextField(
        description="The description of the task.",
        default="",
    )

    task_name = fields.CharField(
        max_length=255,
        description="The task to be executed. This is the path to the task function.",
    )

    task_args = fields.TextField(
        description="The args to be passed to the task function.",
    )

    task_kwargs = fields.TextField(
        description="The kwargs to be passed to the task function.",
    )

    labels = fields.TextField(
        description="The labels to be used to filter the tasks.",
    )

    cron = fields.CharField(
        max_length=255,
        description="The cron expression to be used to schedule the task.",
        default=None,
        null=True,
    )

    time = fields.DatetimeField(
        description="The time to be used to schedule the task.",
        default=None,
        null=True,
    )

    last_run_at = fields.DatetimeField(
        null=True,
        default="1970-01-01 00:00:00",
        description="The last time the task was run.",
    )
    total_run_count = fields.IntField(
        default=0,
        description="The total number of times the task has been run.",
    )

    enabled = fields.SmallIntField(
        default=1,
        description="Whether the task is enabled.",
    )

    schedule_id = fields.CharField(
        max_length=255,
        description="The id of the schedule.",
        default=lambda: str(uuid.uuid4().hex),
        unique=True,
    )

    def to_taskiq_schedule_task(self) -> ScheduledTask:
        base_data = {
            "task_name": self.task_name,
            "args": json.loads(self.task_args.encode()),
            "kwargs": json.loads(self.task_kwargs.encode()),
            "labels": json.loads(self.labels.encode()),
            "schedule_id": self.schedule_id,
        }

        if self.cron:
            base_data["cron"] = self.cron
            return ScheduledTask.model_validate(base_data)

        if self.time:
            base_data["time"] = self.time
            return ScheduledTask.model_validate(base_data)

        raise RuntimeError("No schedule found")



import typing as t
from datetime import datetime

import orjson as json
from taskiq import ScheduledTask, ScheduleSource
from taskiq.abc.serializer import TaskiqSerializer
from taskiq.utils import maybe_awaitable
from tortoise import BaseDBAsyncClient, Tortoise
from tortoise.expressions import F
from unfazed.conf import UnfazedSettings, settings
from unfazed.utils import import_string

from unfazed_taskiq.contrib.scheduler import models as m
from unfazed_taskiq.logger import log


class TortoiseScheduleSource(ScheduleSource):
    def __init__(
        self,
        db_alias: str = "default",
        schedule_alias: str = "default",
        startup_handlers: t.List[str] = [],
        shutdown_handlers: t.List[str] = [],
        serializer: t.Optional[TaskiqSerializer] = None,
    ) -> None:
        """
        Initialize the TortoiseScheduleSource.

        :param db_alias: The alias of the database connection to use.
        :param schedule_alias: The alias of the schedule to use.
        :param startup_handlers: A list of handlers to execute during startup.
        :param shutdown_handlers: A list of handlers to execute during shutdown.
        :param serializer: The serializer to use.
        """
        unfazed_settings: UnfazedSettings = settings["UNFAZED_SETTINGS"]

        if unfazed_settings.DATABASE is None:
            raise RuntimeError("DATABASE is not set in UNFAZED_SETTINGS")

        if db_alias not in unfazed_settings.DATABASE.connections:
            raise RuntimeError(f"Database connection {db_alias} not found")
        self._alias: str = db_alias
        self.schedule_alias: str = schedule_alias
        self.startup_handlers: t.List[str] = startup_handlers
        self.shutdown_handlers: t.List[str] = shutdown_handlers

    async def startup(self) -> None:
        """Action to execute during startup."""
        for handler in self.startup_handlers:
            handler_cls = import_string(handler)
            await maybe_awaitable(handler_cls())

        if not Tortoise._inited:
            raise RuntimeError("Tortoise is not initialized")

        self.alias: BaseDBAsyncClient = Tortoise.get_connection(self._alias)

        log.info("TortoiseScheduleSource startup")

    async def shutdown(self) -> None:
        """Actions to execute during shutdown."""
        for handler in self.shutdown_handlers:
            handler_cls = import_string(handler)
            await maybe_awaitable(handler_cls())

        log.info("MysqlScheduleSource shutdown")

    async def get_schedules(self) -> t.List["ScheduledTask"]:
        """Get list of taskiq schedules."""

        schedules = await m.PeriodicTask.filter(
            enabled=1, schedule_alias=self.schedule_alias
        ).using_db(self.alias)
        return [schedule.to_taskiq_schedule_task() for schedule in schedules]

    async def add_schedule(
        self,
        schedule: "ScheduledTask",
    ) -> None:
        """
        Add a new schedule.

        This function is used to add new schedules.
        It's a convenient helper for people who want to add new schedules
        for the current source.

        As an example, if your source works with a database,
        you may want to add new rows to the table.

        Note that this function may do nothing.

        :param schedule: schedule to add.
        """

        schedule_id = schedule.schedule_id

        if (
            await m.PeriodicTask.filter(schedule_id=schedule_id)
            .using_db(self.alias)
            .exists()
        ):
            raise RuntimeError(f"Schedule {schedule_id} already exists")

        pt = m.PeriodicTask(
            task_name=schedule.task_name,
            task_args=json.dumps(schedule.args).decode(),
            task_kwargs=json.dumps(schedule.kwargs).decode(),
            labels=json.dumps(schedule.labels).decode(),
            schedule_id=schedule_id,
            schedule_alias=self.schedule_alias,
        )

        if schedule.cron is not None:
            pt.cron = schedule.cron
        elif schedule.time is not None:
            pt.time = schedule.time
        else:
            raise RuntimeError("No schedule found")

        await pt.save(using_db=self.alias)

    async def delete_schedule(self, schedule_id: str) -> None:
        """
        Method to delete schedule by id.

        This is useful for schedule cancelation.

        :param schedule_id: id of schedule to delete.
        """

        await (
            m.PeriodicTask.filter(schedule_id=schedule_id)
            .using_db(self.alias)
            .update(enabled=0)
        )

    async def pre_send(  # type: ignore
        self, task: "ScheduledTask"
    ) -> t.Union[  # type: ignore
        None, "t.CoroutineType[t.Any, t.Any, None]", t.Coroutine[t.Any, t.Any, None]
    ]:
        """
        Actions to execute before task will be sent to broker.

        This method may raise ScheduledTaskCancelledError.
        This cancels the task execution.

        :param task: task that will be sent
        """
        await (
            m.PeriodicTask.filter(schedule_id=task.schedule_id)
            .using_db(self.alias)
            .update(last_run_at=datetime.now())
        )

    async def post_send(  # type: ignore
        self, task: "ScheduledTask"
    ) -> t.Union[  # type: ignore
        None, "t.CoroutineType[t.Any, t.Any, None]", t.Coroutine[t.Any, t.Any, None]
    ]:
        """
        Actions to execute after task was sent to broker.

        :param task: task that just have sent
        """

        await (
            m.PeriodicTask.filter(schedule_id=task.schedule_id)
            .using_db(self.alias)
            .update(total_run_count=F("total_run_count") + 1)
        )
